@@ -22,7 +22,7 @@
  * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
  * file for a list of people on the GTK+ Team.  See the ChangeLog
  * files for a list of changes.  These files are distributed with
- * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
+ * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
  */
 
 /*
@@ -31,7 +31,7 @@
  *
  * The entry is now always invisible, uses secure memory methods to
  * allocate the text memory, and all potentially dangerous methods
- * (copy & paste, popup, etc.) have been removed.
+ * (copy, popup, etc.) have been removed.
  */
 
 #include <stdlib.h>
@@ -42,11 +42,6 @@
 #include "gtksecentry.h"
 #include "memory.h"
 
-#ifndef _
-#  include <libintl.h>
-#  define _(x) gettext(x)
-#endif
-
 #define MIN_SECURE_ENTRY_WIDTH  150
 #define DRAW_TIMEOUT     20
 #define INNER_BORDER     2
@@ -62,6 +57,7 @@ enum {
     MOVE_CURSOR,
     INSERT_AT_CURSOR,
     DELETE_FROM_CURSOR,
+    PASTE_CLIPBOARD,
     LAST_SIGNAL
 };
 
@@ -75,7 +71,8 @@ enum {
     PROP_ACTIVATES_DEFAULT,
     PROP_WIDTH_CHARS,
     PROP_SCROLL_OFFSET,
-    PROP_TEXT
+    PROP_TEXT,
+    PROP_EDITING_CANCELED
 };
 
 static guint signals[LAST_SIGNAL] = { 0 };
@@ -179,6 +176,7 @@ static void gtk_secure_entry_real_activate(GtkSecureEntry * entry);
 static void gtk_secure_entry_keymap_direction_changed(GdkKeymap * keymap,
 						      GtkSecureEntry *
 						      entry);
+
 /* IM Context Callbacks
  */
 static void gtk_secure_entry_commit_cb(GtkIMContext * context,
@@ -199,6 +197,13 @@ static gboolean gtk_secure_entry_delete_surrounding_cb(GtkIMContext *
 
 /* Internal routines
  */
+
+static void begin_change (GtkSecureEntry *entry);
+static void end_change (GtkSecureEntry *entry);
+static void emit_changed (GtkSecureEntry *entry);
+static void gtk_secure_entry_paste (GtkSecureEntry *entry, GdkAtom selection);
+static void gtk_secure_entry_paste_clipboard (GtkSecureEntry *entry);
+
 static void gtk_secure_entry_enter_text(GtkSecureEntry * entry,
 					const gchar * str);
 static void gtk_secure_entry_set_positions(GtkSecureEntry * entry,
@@ -282,7 +287,7 @@ secentry_malloc (GMALLOC_SIZE size)
     else
 	p = (gpointer) malloc(size);
     if (!p)
-	g_error("could not allocate %ld bytes", size);
+      g_error("could not allocate %lu bytes", (unsigned long)size);
 
     return p;
 }
@@ -428,32 +433,34 @@ gtk_secure_entry_class_init(GtkSecureEntryClass * class)
     class->move_cursor = gtk_secure_entry_move_cursor;
     class->insert_at_cursor = gtk_secure_entry_insert_at_cursor;
     class->delete_from_cursor = gtk_secure_entry_delete_from_cursor;
+    class->paste_clipboard = gtk_secure_entry_paste_clipboard;
     class->activate = gtk_secure_entry_real_activate;
 
+    g_object_class_override_property (gobject_class,
+                                      PROP_EDITING_CANCELED,
+                                      "editing-canceled");
+
     g_object_class_install_property(gobject_class,
 				    PROP_CURSOR_POSITION,
 				    g_param_spec_int("cursor_position",
-						     _("Cursor Position"),
-						     _
-						     ("The current position of the insertion cursor in chars"),
+						     "Cursor Position",
+						     "The current position of the insertion cursor in chars",
 						     0, MAX_SIZE, 0,
 						     G_PARAM_READABLE));
 
     g_object_class_install_property(gobject_class,
 				    PROP_SELECTION_BOUND,
 				    g_param_spec_int("selection_bound",
-						     _("Selection Bound"),
-						     _
-						     ("The position of the opposite end of the selection from the cursor in chars"),
+						     "Selection Bound",
+						     "The position of the opposite end of the selection from the cursor in chars",
 						     0, MAX_SIZE, 0,
 						     G_PARAM_READABLE));
 
     g_object_class_install_property(gobject_class,
 				    PROP_MAX_LENGTH,
 				    g_param_spec_int("max_length",
-						     _("Maximum length"),
-						     _
-						     ("Maximum number of characters for this entry. Zero if no maximum"),
+						     "Maximum length",
+						     "Maximum number of characters for this entry. Zero if no maximum",
 						     0, MAX_SIZE, 0,
 						     G_PARAM_READABLE |
 						     G_PARAM_WRITABLE));
@@ -461,9 +468,8 @@ gtk_secure_entry_class_init(GtkSecureEntryClass * class)
     g_object_class_install_property(gobject_class,
 				    PROP_HAS_FRAME,
 				    g_param_spec_boolean("has_frame",
-							 _("Has Frame"),
-							 _
-							 ("FALSE removes outside bevel from entry"),
+							 "Has Frame",
+							 "FALSE removes outside bevel from entry",
 							 TRUE,
 							 G_PARAM_READABLE |
 							 G_PARAM_WRITABLE));
@@ -471,10 +477,8 @@ gtk_secure_entry_class_init(GtkSecureEntryClass * class)
     g_object_class_install_property(gobject_class,
 				    PROP_INVISIBLE_CHAR,
 				    g_param_spec_unichar("invisible_char",
-							 _
-							 ("Invisible character"),
-							 _
-							 ("The character to use when masking entry contents (in \"password mode\")"),
+							 "Invisible character",
+							 "The character to use when masking entry contents (in \"password mode\")",
 							 '*',
 							 G_PARAM_READABLE |
 							 G_PARAM_WRITABLE));
@@ -483,16 +487,14 @@ gtk_secure_entry_class_init(GtkSecureEntryClass * class)
 				    PROP_ACTIVATES_DEFAULT,
 				    g_param_spec_boolean
 				    ("activates_default",
-				     _("Activates default"),
-				     _
-				     ("Whether to activate the default widget (such as the default button in a dialog) when Enter is pressed"),
+				     "Activates default",
+				     "Whether to activate the default widget (such as the default button in a dialog) when Enter is pressed",
 				     FALSE,
 				     G_PARAM_READABLE | G_PARAM_WRITABLE));
     g_object_class_install_property(gobject_class, PROP_WIDTH_CHARS,
 				    g_param_spec_int("width_chars",
-						     _("Width in chars"),
-						     _
-						     ("Number of characters to leave space for in the entry"),
+						     "Width in chars",
+						     "Number of characters to leave space for in the entry",
 						     -1, G_MAXINT, -1,
 						     G_PARAM_READABLE |
 						     G_PARAM_WRITABLE));
@@ -500,18 +502,16 @@ gtk_secure_entry_class_init(GtkSecureEntryClass * class)
     g_object_class_install_property(gobject_class,
 				    PROP_SCROLL_OFFSET,
 				    g_param_spec_int("scroll_offset",
-						     _("Scroll offset"),
-						     _
-						     ("Number of pixels of the entry scrolled off the screen to the left"),
+						     "Scroll offset",
+						     "Number of pixels of the entry scrolled off the screen to the left",
 						     0, G_MAXINT, 0,
 						     G_PARAM_READABLE));
 
     g_object_class_install_property(gobject_class,
 				    PROP_TEXT,
 				    g_param_spec_string("text",
-							_("Text"),
-							_
-							("The contents of the entry"),
+							"Text",
+							"The contents of the entry",
 							"",
 							G_PARAM_READABLE |
 							G_PARAM_WRITABLE));
@@ -554,6 +554,15 @@ gtk_secure_entry_class_init(GtkSecureEntryClass * class)
 		     _gtk_marshal_VOID__ENUM_INT, G_TYPE_NONE, 2,
 		     GTK_TYPE_DELETE_TYPE, G_TYPE_INT);
 
+    signals[PASTE_CLIPBOARD] =
+      g_signal_new ("paste-clipboard",
+                    G_OBJECT_CLASS_TYPE (gobject_class),
+                    G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
+                    G_STRUCT_OFFSET (GtkEntryClass, paste_clipboard),
+                    NULL, NULL,
+                    _gtk_marshal_VOID__VOID,
+                    G_TYPE_NONE, 0);
+
     /*
      * Key bindings
      */
@@ -665,6 +674,15 @@ gtk_secure_entry_class_init(GtkSecureEntryClass * class)
 				 GDK_CONTROL_MASK, "delete_from_cursor", 2,
 				 G_TYPE_ENUM, GTK_DELETE_WORD_ENDS,
 				 G_TYPE_INT, -1);
+
+    /* Clipboard - only pasting of course.  */
+    /* Ist does not work right now.  Depending on the GTK version we
+       see segv due to "property added after class was intialized".  */
+    /* gtk_binding_entry_add_signal (binding_set, GDK_v, GDK_CONTROL_MASK, */
+    /*                               "paste-clipboard", 0); */
+    /* gtk_binding_entry_add_signal (binding_set, GDK_Insert, GDK_SHIFT_MASK, */
+    /*                               "paste-clipboard", 0); */
+
 }
 
 static void
@@ -720,6 +738,14 @@ gtk_secure_entry_set_property(GObject * object,
 	gtk_secure_entry_set_text(entry, g_value_get_string(value));
 	break;
 
+    case PROP_EDITING_CANCELED:
+      /* We may want to implement the GtkCelllEditable methods.
+         However it is only used by GtkTreeView and we don't use that
+         here.  We provide this property only to get rid of a Gtk+
+         runtime warning.  */
+        entry->editing_canceled = g_value_get_boolean (value);
+        break;
+
     case PROP_SCROLL_OFFSET:
     case PROP_CURSOR_POSITION:
     default:
@@ -763,6 +789,9 @@ gtk_secure_entry_get_property(GObject * object,
     case PROP_TEXT:
 	g_value_set_string(value, gtk_secure_entry_get_text(entry));
 	break;
+    case PROP_EDITING_CANCELED:
+        g_value_set_boolean (value, entry->editing_canceled);
+        break;
 
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
@@ -832,13 +861,11 @@ static void
 gtk_secure_entry_realize(GtkWidget * widget)
 {
     GtkSecureEntry *entry;
-    GtkEditable *editable;
     GdkWindowAttr attributes;
     gint attributes_mask;
 
     GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);
     entry = GTK_SECURE_ENTRY(widget);
-    editable = GTK_EDITABLE(widget);
 
     attributes.window_type = GDK_WINDOW_CHILD;
 
@@ -1166,6 +1193,18 @@ gtk_secure_entry_button_press(GtkWidget * widget, GdkEventButton * event)
 
 	return TRUE;
     }
+    else if (event->button == 2) {
+        switch (event->type) {
+        case GDK_BUTTON_PRESS:
+            entry->insert_pos = tmp_pos;
+            gtk_secure_entry_paste (entry, GDK_SELECTION_PRIMARY);
+            return TRUE;
+
+        default:
+            break;
+        }
+
+    }
 
     return FALSE;
 }
@@ -1646,7 +1685,7 @@ gtk_secure_entry_real_insert_text(GtkEditable * editable,
 
     gtk_secure_entry_recompute(entry);
 
-    g_signal_emit_by_name(editable, "changed");
+    emit_changed (entry);
     g_object_notify(G_OBJECT(editable), "text");
 }
 
@@ -1687,7 +1726,7 @@ gtk_secure_entry_real_delete_text(GtkEditable * editable,
 
 	gtk_secure_entry_recompute(entry);
 
-	g_signal_emit_by_name(editable, "changed");
+	emit_changed (entry);
 	g_object_notify(G_OBJECT(editable), "text");
     }
 }
@@ -1860,21 +1899,115 @@ gtk_secure_entry_delete_from_cursor(GtkSecureEntry * entry,
 }
 
 static void
-gtk_secure_entry_delete_cb(GtkSecureEntry * entry)
+begin_change (GtkSecureEntry *entry)
 {
-    GtkEditable *editable = GTK_EDITABLE(entry);
-    gint start, end;
-
-    if (gtk_editable_get_selection_bounds(editable, &start, &end))
-      gtk_editable_delete_text(editable, start, end);
+  entry->change_count++;
 }
 
 static void
-gtk_secure_entry_toggle_overwrite(GtkSecureEntry * entry)
+end_change (GtkSecureEntry *entry)
 {
-    entry->overwrite_mode = !entry->overwrite_mode;
+  GtkEditable *editable = GTK_EDITABLE (entry);
+
+  g_return_if_fail (entry->change_count > 0);
+
+  entry->change_count--;
+
+  if (entry->change_count == 0)
+    {
+       if (entry->real_changed)
+         {
+           g_signal_emit_by_name (editable, "changed");
+           entry->real_changed = FALSE;
+         }
+    }
 }
 
+static void
+emit_changed (GtkSecureEntry *entry)
+{
+  GtkEditable *editable = GTK_EDITABLE (entry);
+
+  if (entry->change_count == 0)
+    g_signal_emit_by_name (editable, "changed");
+  else
+    entry->real_changed = TRUE;
+}
+
+
+static void
+paste_received (GtkClipboard *clipboard,
+		const gchar  *text,
+		gpointer      data)
+{
+  GtkSecureEntry *entry = GTK_SECURE_ENTRY (data);
+  GtkEditable *editable = GTK_EDITABLE (entry);
+
+  if (entry->button == 2)
+    {
+      gint pos, start, end;
+
+      pos = entry->insert_pos;
+      gtk_editable_get_selection_bounds (editable, &start, &end);
+      if (!((start <= pos && pos <= end) || (end <= pos && pos <= start)))
+	gtk_editable_select_region (editable, pos, pos);
+    }
+
+  if (text)
+    {
+      gint pos, start, end;
+      gint length = -1;
+
+      begin_change (entry);
+      g_object_freeze_notify (G_OBJECT (entry));
+      if (gtk_editable_get_selection_bounds (editable, &start, &end))
+        gtk_editable_delete_text (editable, start, end);
+
+      pos = entry->current_pos;
+      gtk_editable_insert_text (editable, text, length, &pos);
+      gtk_editable_set_position (editable, pos);
+      g_object_thaw_notify (G_OBJECT (entry));
+      end_change (entry);
+    }
+
+  g_object_unref (entry);
+}
+
+
+static void
+gtk_secure_entry_paste (GtkSecureEntry *entry, GdkAtom selection)
+{
+  g_object_ref (entry);
+  gtk_clipboard_request_text (gtk_widget_get_clipboard (GTK_WIDGET (entry),
+                                                        selection),
+			      paste_received, entry);
+}
+
+
+static void
+gtk_secure_entry_paste_clipboard (GtkSecureEntry *entry)
+{
+  gtk_secure_entry_paste (entry, GDK_SELECTION_CLIPBOARD);
+}
+
+
+
+/* static void */
+/* gtk_secure_entry_delete_cb(GtkSecureEntry * entry) */
+/* { */
+/*     GtkEditable *editable = GTK_EDITABLE(entry); */
+/*     gint start, end; */
+
+/*     if (gtk_editable_get_selection_bounds(editable, &start, &end)) */
+/*       gtk_editable_delete_text(editable, start, end); */
+/* } */
+
+/* static void */
+/* gtk_secure_entry_toggle_overwrite(GtkSecureEntry * entry) */
+/* { */
+/*     entry->overwrite_mode = !entry->overwrite_mode; */
+/* } */
+
 static void
 gtk_secure_entry_real_activate(GtkSecureEntry * entry)
 {
@@ -2670,14 +2803,14 @@ gtk_secure_entry_new(void)
  *   be clamped to the range 0-65536.
  *
  * Creates a new #GtkSecureEntry widget with the given maximum length.
- * 
+ *
  * Note: the existence of this function is inconsistent
  * with the rest of the GTK+ API. The normal setup would
  * be to just require the user to make an extra call
  * to gtk_secure_entry_set_max_length() instead. It is not
  * expected that this function will be removed, but
  * it would be better practice not to use it.
- * 
+ *
  * Return value: a new #GtkSecureEntry.
  **/
 GtkWidget *
@@ -2750,7 +2883,7 @@ gtk_secure_entry_set_position(GtkSecureEntry * entry, gint position)
  * gtk_secure_entry_set_invisible_char:
  * @entry: a #GtkSecureEntry
  * @ch: a Unicode character
- * 
+ *
  * Sets the character to use in place of the actual text when
  * gtk_secure_entry_set_visibility() has been called to set text visibility
  * to %FALSE. i.e. this is the character used in "password mode" to
@@ -2758,7 +2891,7 @@ gtk_secure_entry_set_position(GtkSecureEntry * entry, gint position)
  * invisible char is an asterisk ('*').  If you set the invisible char
  * to 0, then the user will get no feedback at all; there will be
  * no text on the screen as they type.
- * 
+ *
  **/
 void
 gtk_secure_entry_set_invisible_char(GtkSecureEntry * entry, gunichar ch)
@@ -2781,7 +2914,7 @@ gtk_secure_entry_set_invisible_char(GtkSecureEntry * entry, gunichar ch)
  * for entries with visisbility set to false. See gtk_secure_entry_set_invisible_char().
  *
  * Return value: the current invisible char, or 0, if the entry does not
- *               show invisible text at all. 
+ *               show invisible text at all.
  **/
 gunichar
 gtk_secure_entry_get_invisible_char(GtkSecureEntry * entry)
@@ -2824,7 +2957,7 @@ gtk_secure_entry_select_region(GtkSecureEntry * entry,
  * @max: the maximum length of the entry, or 0 for no maximum.
  *   (other than the maximum length of entries.) The value passed in will
  *   be clamped to the range 0-65536.
- * 
+ *
  * Sets the maximum allowed length of the contents of the widget. If
  * the current contents are longer than the given length, then they
  * will be truncated to fit.
@@ -2874,7 +3007,7 @@ gtk_secure_entry_get_max_length(GtkSecureEntry * entry)
  * (For experts: if @setting is %TRUE, the entry calls
  * gtk_window_activate_default() on the window containing the entry, in
  * the default handler for the "activate" signal.)
- * 
+ *
  **/
 void
 gtk_secure_entry_set_activates_default(GtkSecureEntry * entry,
@@ -2892,9 +3025,9 @@ gtk_secure_entry_set_activates_default(GtkSecureEntry * entry,
 /**
  * gtk_secure_entry_get_activates_default:
  * @entry: a #GtkSecureEntry
- * 
+ *
  * Retrieves the value set by gtk_secure_entry_set_activates_default().
- * 
+ *
  * Return value: %TRUE if the entry will activate the default widget
  **/
 gboolean
@@ -2915,7 +3048,7 @@ gtk_secure_entry_get_activates_default(GtkSecureEntry * entry)
  * <emphasis>request</emphasis>, the size can still be affected by
  * how you pack the widget into containers. If @n_chars is -1, the
  * size reverts to the default entry size.
- * 
+ *
  **/
 void
 gtk_secure_entry_set_width_chars(GtkSecureEntry * entry, gint n_chars)
@@ -2932,9 +3065,9 @@ gtk_secure_entry_set_width_chars(GtkSecureEntry * entry, gint n_chars)
 /**
  * gtk_secure_entry_get_width_chars:
  * @entry: a #GtkSecureEntry
- * 
+ *
  * Gets the value set by gtk_secure_entry_set_width_chars().
- * 
+ *
  * Return value: number of chars to request space for, or negative if unset
  **/
 gint
@@ -2949,7 +3082,7 @@ gtk_secure_entry_get_width_chars(GtkSecureEntry * entry)
  * gtk_secure_entry_set_has_frame:
  * @entry: a #GtkSecureEntry
  * @setting: new value
- * 
+ *
  * Sets whether the entry has a beveled frame around it.
  **/
 void
@@ -2970,9 +3103,9 @@ gtk_secure_entry_set_has_frame(GtkSecureEntry * entry, gboolean setting)
 /**
  * gtk_secure_entry_get_has_frame:
  * @entry: a #GtkSecureEntry
- * 
+ *
  * Gets the value set by gtk_secure_entry_set_has_frame().
- * 
+ *
  * Return value: whether the entry has a beveled frame
  **/
 gboolean
@@ -2987,7 +3120,7 @@ gtk_secure_entry_get_has_frame(GtkSecureEntry * entry)
 /**
  * gtk_secure_entry_get_layout:
  * @entry: a #GtkSecureEntry
- * 
+ *
  * Gets the #PangoLayout used to display the entry.
  * The layout is useful to e.g. convert text positions to
  * pixel positions, in combination with gtk_secure_entry_get_layout_offsets().
@@ -2998,7 +3131,7 @@ gtk_secure_entry_get_has_frame(GtkSecureEntry * entry)
  * gtk_secure_entry_layout_index_to_text_index() and
  * gtk_secure_entry_text_index_to_layout_index() are needed to convert byte
  * indices in the layout to byte indices in the entry contents.
- * 
+ *
  * Return value: the #PangoLayout for this entry
  **/
 PangoLayout *
@@ -3018,12 +3151,12 @@ gtk_secure_entry_get_layout(GtkSecureEntry * entry)
  * gtk_secure_entry_layout_index_to_text_index:
  * @entry: a #GtkSecureEntry
  * @layout_index: byte index into the entry layout text
- * 
+ *
  * Converts from a position in the entry contents (returned
  * by gtk_secure_entry_get_text()) to a position in the
  * entry's #PangoLayout (returned by gtk_secure_entry_get_layout(),
  * with text retrieved via pango_layout_get_text()).
- * 
+ *
  * Return value: byte index into the entry contents
  **/
 gint
@@ -3055,11 +3188,11 @@ gtk_secure_entry_layout_index_to_text_index(GtkSecureEntry * entry,
  * gtk_secure_entry_text_index_to_layout_index:
  * @entry: a #GtkSecureEntry
  * @text_index: byte index into the entry contents
- * 
+ *
  * Converts from a position in the entry's #PangoLayout(returned by
  * gtk_secure_entry_get_layout()) to a position in the entry contents
  * (returned by gtk_secure_entry_get_text()).
- * 
+ *
  * Return value: byte index into the entry layout text
  **/
 gint
@@ -3097,7 +3230,7 @@ gtk_secure_entry_text_index_to_layout_index(GtkSecureEntry * entry,
  * Also useful to convert mouse events into coordinates inside the
  * #PangoLayout, e.g. to take some action if some part of the entry text
  * is clicked.
- * 
+ *
  * Note that as the user scrolls around in the entry the offsets will
  * change; you'll need to connect to the "notify::scroll_offset"
  * signal to track this. Remember when using the #PangoLayout
@@ -3108,7 +3241,7 @@ gtk_secure_entry_text_index_to_layout_index(GtkSecureEntry * entry,
  * gtk_secure_entry_layout_index_to_text_index() and
  * gtk_secure_entry_text_index_to_layout_index() are needed to convert byte
  * indices in the layout to byte indices in the entry contents.
- * 
+ *
  **/
 void
 gtk_secure_entry_get_layout_offsets(GtkSecureEntry * entry,
@@ -3134,13 +3267,13 @@ gtk_secure_entry_get_layout_offsets(GtkSecureEntry * entry,
 
 /* Quick hack of a popup menu
  */
-static void
-activate_cb(GtkWidget * menuitem, GtkSecureEntry * entry)
-{
-    const gchar *signal =
-	g_object_get_data(G_OBJECT(menuitem), "gtk-signal");
-    g_signal_emit_by_name(entry, signal);
-}
+/* static void */
+/* activate_cb(GtkWidget * menuitem, GtkSecureEntry * entry) */
+/* { */
+/*   const gchar *asignal = */
+/*     g_object_get_data(G_OBJECT(menuitem), "gtk-signal"); */
+/*   g_signal_emit_by_name(entry, asignal); */
+/* } */
 
 
 static gboolean
@@ -3152,13 +3285,13 @@ gtk_secure_entry_mnemonic_activate(GtkWidget * widget,
 }
 
 
-static void
-unichar_chosen_func(const char *text, gpointer data)
-{
-    GtkSecureEntry *entry = GTK_SECURE_ENTRY(data);
+/* static void */
+/* unichar_chosen_func(const char *text, gpointer data) */
+/* { */
+/*     GtkSecureEntry *entry = GTK_SECURE_ENTRY(data); */
 
-    gtk_secure_entry_enter_text(entry, text);
-}
+/*     gtk_secure_entry_enter_text(entry, text); */
+/* } */
 
 /* We display the cursor when
  *
